
/*  
 *  hostNameCache.cpp,v 1.11 2001/10/22 03:15:44 jhill Exp
 *
 *                              
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *                                  
 *  Copyright, 1986, The Regents of the University of California.
 *                                  
 *           
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#define epicsAssertAuthor "Jeff Hill johill@lanl.gov"

#include <string.h>

#include "iocinf.h"
#include "hostNameCache.h"

#if defined ( _MSC_VER )
#   pragma warning ( push )
#   pragma warning ( disable: 4660 )
#endif

template class tsFreeList < hostNameCache, 16 >;

#if defined ( _MSC_VER )
#   pragma warning ( pop )
#endif

tsFreeList < hostNameCache, 16 > hostNameCache::freeList;
epicsMutex hostNameCache::freeListMutex;

hostNameCache::hostNameCache ( const osiSockAddr &addr, ipAddrToAsciiEngine &engine ) :
    ipAddrToAsciiAsynchronous ( addr ),
    ioComplete ( false )
{
    this->ioInitiate ( engine );
}

void hostNameCache::destroy ()
{
    delete this;
}

hostNameCache::~hostNameCache ()
{
}

void hostNameCache::ioCompletionNotify ( const char *pHostNameIn )
{
    if ( ! this->ioComplete ) {
        strncpy ( this->hostNameBuf, pHostNameIn, sizeof ( this->hostNameBuf ) );
        this->hostNameBuf[ sizeof ( this->hostNameBuf ) - 1 ] = '\0';
    }
}

void hostNameCache::hostName ( char *pBuf, unsigned bufSize ) const
{
    if ( this->ioComplete ) {
        strncpy ( pBuf, this->hostNameBuf, bufSize);
        pBuf [ bufSize - 1u ] = '\0';
    }
    else {
        osiSockAddr tmpAddr = this->address ();
        sockAddrToDottedIP ( &tmpAddr.sa, pBuf, bufSize );
    }
}

void * hostNameCache::operator new ( size_t size )
{
    return hostNameCache::freeList.allocate ( size );
}

void hostNameCache::operator delete ( void *pCadaver, size_t size )
{
    hostNameCache::freeList.release ( pCadaver, size );
}
